;; Generated package description from jgraph-mode.el  -*- no-byte-compile: t -*-
(define-package "jgraph-mode" "1.1.0.20201129725" "Major mode for Jgraph files" '((cl-lib "0.5")) :url "https://elpa.gnu.org/packages/jgraph-mode.html" :keywords '("tex" "wp") :authors '(("Stefan Monnier" . "monnier@iro.umontreal.ca")) :maintainer '("Stefan Monnier" . "monnier@iro.umontreal.ca"))
